#!/usr/bin/env python
# Copyright (c) 2006 ActiveState Software Inc.

r"""A command-line interface for managing UDL (User-Defined Language)
files for Komodo.
"""

import os
from os.path import basename, dirname, join, exists, abspath
import sys
import re
import getopt
import stat
import string
import logging
from glob import glob
from pprint import pprint

from ludditelib import cmdln

from ludditelib import parser, gen, constants
from ludditelib.common import LudditeError, __version__


log = logging.getLogger("luddite")



#---- luddite logic

def get_build_dir(lang):
    return join("build", lang)


def compile(udl_path, skel=False, guid=None, guid_from_lang=None, ext=None,
            force=False):
    """Compile the given .udl file to Komodo language resources.
    
    One of "guid" or "guid_from_lang" can be given to specify the XPCOM
    language service GUID. If neither is given then a new one will be
    generated.
    """
    # Dev Notes:
    # - Compiling builds into build/$languange/... and creates:
    #       ${language}.lexres                  lexer resource
    # - If skel is True then also build skeletons for:
    #       ko${language}_UDL_Language.py       language component
    #       ${language}.${ext}                  empty Komodo template
    # - Add support for other options that Eric had in the preceding
    #   luddite.py?
    log.debug("compile('%s', ...)", udl_path)

    # Clean up after PLY. It leaves some turds that can break subsequent
    # parsing if the parser source is changed.
    turds = ["parser.out", "parsetab.py", "parsetab.pyc"]
    for turd in turds:
        if exists(turd):
            log.debug("remove `%s'", turd)
            os.remove(turd)

    # Parse and load the UDL definition.
    parse_tree = parse(udl_path)
    if parse_tree is None:
        raise LudditeError("parse failed");
    mainObj = gen.MainObj()
    analyzer = gen.Analyzer(mainObj)
    analyzer.processTree(parse_tree)
    mainObj.calcUniqueStates()

    #XXX Grr. Crappy error handling again. This should be changed to
    #    raise a LudditeError if the semanticCheck fails then just call:
    #       analyzer.semanticCheck()
    if analyzer.semanticCheck() is None:
        return 1

    def raise_force_error(path):
        raise LudditeError("`%s' already exists: use "
                           "-f|--force option to allow it to be "
                           "overwritted" % path)

    # Make sure don't trounce files before generating outputs and setup
    # output dir.
    lang_name = mainObj.languageName
    safe_lang_name = mainObj.safeLangName
    build_dir = get_build_dir(lang_name)
    if not exists(build_dir):
        log.debug("mkdir `%s'", build_dir)
        os.makedirs(build_dir)

    # Generate all outputs.
    lexres_path = join(build_dir, "lexers", safe_lang_name+".lexres")
    if exists(lexres_path):
        if not force: raise_force_error(lexres_path)
        log.debug("rm `%s'", lexres_path)
        os.remove(lexres_path)
    if not exists(dirname(lexres_path)):
        os.makedirs(dirname(lexres_path))
    log.info("create lexres `%s'", lexres_path)
    mainObj.dumpAsTable(constants.vals, lexres_path)

    if skel:
        lang_service_path \
            = join(build_dir, "components", "ko%s_UDL_Language.py" % safe_lang_name)
        if exists(lang_service_path):
            if not force: raise_force_error(lang_service_path)
            log.debug("rm `%s'", lang_service_path)
            os.remove(lang_service_path)
        if not exists(dirname(lang_service_path)):
            os.makedirs(dirname(lang_service_path))
        log.info("create lang service `%s'", lang_service_path)
        if guid is not None:
            assert _guid_pat.match(guid)
            guid = _norm_guid(guid)
        elif guid_from_lang:
            try:
                guid = guid_from_lang[mainObj.languageName]
            except KeyError:
                raise LudditeError("could not find `%s' in GUIDs text file"
                                   % mainObj.languageName)
        else:
            log.warn("generating new GUID for `%s' language service: it is "
                     "recommended that you use the -g|--guid option to ensure "
                     "a constant GUID from build to build",
                     mainObj.languageName)
            guid = _generate_guid()
        mainObj.dumpLanguageService(lang_service_path, guid, ext=ext)

        if not ext:
            log.warn("no file extension was given: no skeleton "
                     "Komodo templates will be created")
        else:
            template_paths = [
                join(build_dir, "templates", "All Languages",
                     safe_lang_name+ext),
                join(build_dir, "templates", "Common", safe_lang_name+ext),
            ]
            for template_path in template_paths:
                if exists(template_path):
                    if not force: raise_force_error(template_path)
                    log.debug("rm `%s'", template_path)
                    os.remove(template_path)
                if not exists(dirname(template_path)):
                    os.makedirs(dirname(template_path))
                log.info("create template `%s'", template_path)
                mainObj.generateKomodoTemplateFile(template_path)

    # Clean up after PLY. It leaves some turds that can break subsequent
    # parsing if the parser source is changed.
    for turd in turds:
        if exists(turd):
            log.debug("remove `%s'", turd)
            os.remove(turd)


def parse(udl_path):
    """Parse the given .udl file.
    
    PLY (i.e. yacc.py and lex.py) are messy. They leave 'parser.out' and
    'parsetab.py' turds in the current directory. Attempting to hack around
    this by cd'ing into the build dir to run. However this breaks the parse
    for a reason I don't understand. It would be good to fix that at some
    point.
    """
    log.debug("parse('%s')", udl_path)
    parse_tree = parser.parse_udl_path(udl_path)

    # Ick. This modules uses a global for an error count.
    if parser.num_errs:
        raise LudditeError("could not parse '%s': %d error(s)"
                           % (udl_path, parser.num_errs))
    return parse_tree
    

def package(language_name, version=None, creator=None,
            name=None, description=None,
            id=None, force=False):
    """Package the (built) resources for the given languages into a Komodo
    extension.
    
        "language_name" is the language name for which to build a package.
            The resources for this language must already have been built via
            "luddite.py compile ...".

    These arguments are optional, but should be specified:
        "version" (optional, default "1.0.0") is the version number for this
            package.
        "creator" (optional, default "Anonymous Coward") is the name of the
            person creating/maintaining this package.

    These arguments are optional and it is generally fine to not specify them
    because they have reasonable defaults:
        "name" (optional) is the name for the extension.
        "description" (optional) is a short description of the extension.
        "id" (optional) is the extension's id -- an internal short string used
            as a key to identify the extension. It is used in the extension's
            install path.
    
    Dev Notes:
    - For now packaging requires a 'zip' executable somewhere on the
      PATH. This *could* be removed (by using Python's zlib) if too
      burdensome.
    """
    log.debug("package(%r, force=%r)", language_name, force)

    build_dir = get_build_dir(language_name)
    if not exists(build_dir):
        raise LudditeError("`%s': the build dir does not exist: you must first "
                           "build the language resources with "
                           "'luddite compile ...'" % build_dir)

    # Determine package info and create the extension's install.rdf.
    if name is None:
        name = language_name + " Language"
    codename = _codename_from_name(name)
    if id is None:
        id = "%s@ActiveState.com" % codename
    if version is None:
        version = "1.0.0"
        log.warn("defaulting 'version' to '%s' (use version option)",
                 version)
    #else:
    #    XXX validate version
    if description is None:
        description = "%s language support for Komodo (UDL-based)" % language_name
    if creator is None:
        creator = "Anonymous Coward"
        log.warn("defaulting 'creator' to '%s' (use creator option)", creator)

    install_rdf_in = join(dirname(constants.__file__), "install.rdf.in")
    log.debug("reading 'install.rdf' template from '%s'", install_rdf_in)
    install_rdf_template = string.Template(open(install_rdf_in, 'r').read())
    install_rdf = install_rdf_template.substitute(
        id=id, name=name, codename=codename, version=version,
        description=description, creator=creator)
    install_rdf_path = join(build_dir, "install.rdf")
    log.info("create `%s'", install_rdf_path)
    fout = open(install_rdf_path, 'w')
    fout.write(install_rdf)
    fout.close()

    # Create the chrome.manifest file, if necessary. (Empty because we don't
    # install skin, locales or chrome, but should still have it).
    chrome_manifest_path = join(build_dir, "chrome.manifest")
    if not exists(chrome_manifest_path):
        log.info("create `%s'", chrome_manifest_path)
        fout = open(chrome_manifest_path, 'w')
        fout.close()

    # Create the xpi.
    xpi_name = "%s-%s-ko.xpi" % (codename, version)
    xpi_path = xpi_name # put in top-level dir for now
    if exists(xpi_path):
        if not force:
            raise LudditeError("`%s' exists: use force option to overwrite"
                               % xpi_path)
        log.debug("rm `%s'", xpi_path)
        os.remove(xpi_path)
    zip_opts = ""
    if not log.isEnabledFor(logging.DEBUG):
        zip_opts += "-q"
    cmd = 'zip -r %s "%s" *' % (zip_opts, abspath(xpi_path))
    _run_in_dir(cmd, build_dir, log.debug)
    log.info("`%s' successfully created", xpi_path)



#---- internal support stuff

_guid_pat = re.compile(r"^{?[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}"
                        "-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}}?$")

def _norm_guid(g):
    if not g.startswith("{"):
        g = "{"+g
    if not g.endswith("}"):
        g += "}"
    return g

def _generate_guid():
    """Return a GUID/UUID in the typical
        {12345678-90ab-cdef-1234-567890abcdef}
    format (*with* the braces).
    """
    guid = None

    # On Windows, try pythoncom.CreateGuid().
    if sys.platform == "win32":
        try:
            import pythoncom
        except ImportError:
            pass
        else:
            guid = str(pythoncom.CreateGuid())

    # Try Ka-Ping Yee's uuid.py module.
    if guid is None:
        from ludditelib import uuid
        guid = str(uuid.uuid4())
    
    # Fallback to PyXPCOM.
    if guid is None:
        from xpcom import components
        uuidGenerator = components.classes["@mozilla.org/uuid-generator;1"] \
            .getService(components.interfaces.nsIUUIDGenerator)
        guid = str(uuidGenerator.generateUUID())

    return guid



class Shell(cmdln.Cmdln):
    """
    luddite: build and package Komodo language extensions

    usage:
        ${name} SUBCOMMAND [ARGS...]
        ${name} help SUBCOMMAND

    Language syntax-highlighting in Komodo requires a lexer. For most of
    Komodo's core supported languages these lexers are written in C++.
    However, as of Komodo 4, you can define custom lexers for languages that
    Komodo doesn't support out of the box. This system is called UDL -- for
    User-Defined Languages.
    
    Luddite is a tool for building and packaging a Komodo language
    extension. The typical process is:
    
    1. Author a .udl file defining syntax highlighting rules for your
       language.

    2. Compile your .udl file and possibly generate skeleton files for
       the Language Service and "New File" template files that will go
       into the extension.

            python luddite.py compile --skel -G udl/foo-mainlex.udl

       Alternatively you may have a hand-edited language service and
       template files. If so, you just need to move them to the
       appropriate build directory.

    4. Package all the bits into a Komodo extension
            python luddite.py package --version 1.0.0 Foo

    For more information on writing .udl files see Komodo's UDL
    documentation.

    ${option_list}
    ${command_list}
    ${help_list}
    """
    name = "luddite"
    helpindent = ' '*2

    @cmdln.option("--ext",
                  help="specify a default file extension for this language")
    @cmdln.option("-g", "--guid", 
                  help="specify an XPCOM component GUID to use, or a path "
                       "to GUIDs text file")
    @cmdln.option("-G", dest="new_guid", action="store_true",
                  help="create a new XPCOM component GUID for this build")
    @cmdln.option("--skel", action="store_true",
                  help="also build skeleton Language Service and template files")
    @cmdln.option("-f", "--force", action="store_true",
                  help="allow build outputs to overwrite existing files")
    def do_compile(self, subcmd, opts, udl_path):
        """${cmd_name}: compile a .udl file into a .lexres file
        
        ${cmd_usage}
        ${cmd_option_list}
        If you specify '--skel' to build a skeleton Language Service
        then you must also specify on of the -G or -g|--guid options.
        The Language Service is a Python file that controls language
        support in Komodo. It requires a unique GUID (for the XPCOM
        component's class id).

        If you are just building the skeleton language service for every
        build (you can get away with this for minimal language support)
        the using a GUIDs text file to ensure the same GUID is used for
        your component from build to build is recommended. This file
        must be of the form (one entry per line):
            <lang> <guid>
        """
        guid = guid_from_lang = None
        if not opts.skel:
            pass
        elif opts.new_guid and opts.guid:
            raise LudditeError("cannot specify both -G and -g|--guid "
                               "options at the same time")
        elif opts.new_guid:
            guid = None
        elif opts.guid:
            if _guid_pat.match(opts.guid):
                guid = _norm_guid(opts.guid)
            else:
                if not exists(opts.guid):
                    raise LudditeError("`%s' is not a GUID and does not "
                                       "exist" % opts.guid)
                guid_from_lang = {} #dict((lang, g) for lang, g in)
                for line in file(opts.guid):
                    if line.startswith("#"): continue
                    if not line.strip(): continue
                    lang, g = line.strip().split(None, 1)
                    guid_from_lang[lang] = _norm_guid(g)
        else:
            raise LudditeError("must specify one of -G or -g|--guid")
        return compile(udl_path, skel=opts.skel, guid=guid, 
                       guid_from_lang=guid_from_lang, ext=opts.ext,
                       force=opts.force)

    def _do_parse(self, subcmd, opts, *udl_paths):
        """${cmd_name}: parse the given .udl file (for debugging)
        
        ${cmd_usage}
        ${cmd_option_list}
        """
        build_dir = get_build_dir(lang="_parse_test_")
        if not exists(build_dir):
            os.makedirs(build_dir)
        for udl_path in udl_paths:
            tree = parse(udl_path, build_dir)
            pprint(tree)
        return

    @cmdln.option("-f", "--force", action="store_true",
                  help="allow build outputs to overwrite existing files")
    @cmdln.option("--id", action="store",
                  help="the internal identifier for this extension (a short string)")
    @cmdln.option("--name", action="store",
                  help="the extension name")
    @cmdln.option("--description", action="store",
                  help="a short description of the extension")
    @cmdln.option("-c", "--creator", action="store",
                  help="the name of the creator/maintainer of this extension")
    @cmdln.option("-V", "--version", action="store",
                  help="the extension version")
    def do_package(self, subcmd, opts, lang):
        """${cmd_name}: package a compiled language into a Komodo extension for
        distribution

        ${cmd_usage}
        ${cmd_option_list}
        You must first have run '${name} compile ...' to build the Komodo
        resources for this language.

        Typical usage should specify the "creator" and "version". The other
        values have reasonable defaults. For example:
        
            ${name} package -c "Santa Claus" --version "2.0.1" Toy
            ${name} package -c "Larry Wall" --version "0.8.0" Perl

        (However, Komodo already has a Perl lexer so creating your own
        UDL-based one for Perl is just for the masochists.)
        """
        return package(lang,
                       version=opts.version,
                       creator=opts.creator,
                       name=opts.name,
                       description=opts.description,
                       id=opts.id,
                       force=opts.force)


# Recipe: pretty_logging (0.1) in C:\trentm\tm\recipes\cookbook
class _PerLevelFormatter(logging.Formatter):
    """Allow multiple format string -- depending on the log level.
    
    A "fmtFromLevel" optional arg is added to the constructor. It can be
    a dictionary mapping a log record level to a format string. The
    usual "fmt" argument acts as the default.
    """
    def __init__(self, fmt=None, datefmt=None, fmtFromLevel=None):
        logging.Formatter.__init__(self, fmt, datefmt)
        if fmtFromLevel is None:
            self.fmtFromLevel = {}
        else:
            self.fmtFromLevel = fmtFromLevel
    def format(self, record):
        record.levelname = record.levelname.lower()
        if record.levelno in self.fmtFromLevel:
            #XXX This is a non-threadsafe HACK. Really the base Formatter
            #    class should provide a hook accessor for the _fmt
            #    attribute. *Could* add a lock guard here (overkill?).
            _saved_fmt = self._fmt
            self._fmt = self.fmtFromLevel[record.levelno]
            try:
                return logging.Formatter.format(self, record)
            finally:
                self._fmt = _saved_fmt
        else:
            return logging.Formatter.format(self, record)

def _setup_logging():
    hdlr = logging.StreamHandler()
    defaultFmt = "%(name)s: %(levelname)s: %(message)s"
    infoFmt = "%(name)s: %(message)s"
    fmtr = _PerLevelFormatter(fmt=defaultFmt,
                              fmtFromLevel={logging.INFO: infoFmt})
    hdlr.setFormatter(fmtr)
    logging.root.addHandler(hdlr)
    log.setLevel(logging.INFO)
    #log.setLevel(logging.DEBUG)

def _rmtree_on_error(rmFunction, filePath, excInfo):
    if excInfo[0] == OSError:
        # presuming because file is read-only
        os.chmod(filePath, 0777)
        rmFunction(filePath)

def _rmtree(dirname):
    import shutil
    shutil.rmtree(dirname, 0, _rmtree_on_error)


# Recipe: run (0.5.3) in C:\trentm\tm\recipes\cookbook
_RUN_DEFAULT_LOGSTREAM = ("RUN", "DEFAULT", "LOGSTREAM")
def __run_log(logstream, msg, *args, **kwargs):
    if not logstream:
        pass
    elif logstream is _RUN_DEFAULT_LOGSTREAM:
        try:
            log
        except NameError:
            pass
        else:
            if hasattr(log, "debug"):
                log.debug(msg, *args, **kwargs)
    else:
        logstream(msg, *args, **kwargs)

def _run(cmd, logstream=_RUN_DEFAULT_LOGSTREAM):
    """Run the given command.

        "cmd" is the command to run
        "logstream" is an optional logging stream on which to log the 
            command. If None, no logging is done. If unspecifed, this 
            looks for a Logger instance named 'log' and logs the command 
            on log.debug().

    Raises OSError is the command returns a non-zero exit status.
    """
    __run_log(logstream, "running '%s'", cmd)
    retval = os.system(cmd)
    if hasattr(os, "WEXITSTATUS"):
        status = os.WEXITSTATUS(retval)
    else:
        status = retval
    if status:
        #TODO: add std OSError attributes or pick more approp. exception
        raise OSError("error running '%s': %r" % (cmd, status))

def _run_in_dir(cmd, cwd, logstream=_RUN_DEFAULT_LOGSTREAM):
    """Run the given command in the given working directory.

        "cmd" is the command to run
        "cwd" is the directory in which the commmand is run.
        "logstream" is an optional logging stream on which to log the 
            command. If None, no logging is done. If unspecifed, this 
            looks for a Logger instance named 'log' and logs the command 
            on log.debug().

    Raises OSError is the command returns a non-zero exit status.
    """
    old_dir = os.getcwd()
    try:
        os.chdir(cwd)
        __run_log(logstream, "running '%s' in '%s'", cmd, cwd)
        _run(cmd, logstream=None)
    finally:
        os.chdir(old_dir)


def _codename_from_name(name):
    """Transform a Komodo extension name to a "code" name, i.e. one that is
    safe for use in a filename and an extension id.
    """
    return re.sub(r'\W', '_', name.lower())



#---- mainline

_g_verbosity_count = 0
def _set_verbosity(option, opt_str, value, parser):
    global _g_verbosity_count, log
    _g_verbosity_count += 1
    if _g_verbosity_count == 1:
        log.setLevel(logging.INFO)
    elif _g_verbosity_count > 1:
        log.setLevel(logging.DEBUG)

if __name__ == "__main__":
    _setup_logging() # defined in recipe:pretty_logging
    try:
        shell = Shell()
        optparser = cmdln.CmdlnOptionParser(shell,
            version=Shell.name+" "+__version__)
        optparser.add_option("-v", "--verbose", 
            action="callback", callback=_set_verbosity,
            help="More verbose output. Repeat for more and more output.")
        retval = shell.main(sys.argv, optparser=optparser)
    except KeyboardInterrupt:
        sys.exit(1)
    except:
        exc_info = sys.exc_info()
        if hasattr(exc_info[0], "__name__"):
            log.error("%s: %s", exc_info[0].__name__, exc_info[1])
        else:  # string exception
            log.error(exc_info[0])
        if log.isEnabledFor(logging.DEBUG):
            import traceback
            print
            traceback.print_exception(*exc_info)
        sys.exit(1)
    else:
        sys.exit(retval)


